/*************************************************************************
 * CPR.C       " Prints 'C' Files with Table of Contents "
 *             "             AND LOTS MORE               "
 *
 *
 * Microsoft C Version 5.1 
 *
 * Note:
 *    Turbo C Version 1.0 or 1.5 is available as CPR31.ARC
 *    in Borland Forum
 *
 * Originally Written by :
 *    Paul Breslin Human Computing Resources Corp.
 *    10 South Mary Street
 *    Toronto, Ontario
 *    Canada, M4Y 1P9
 *
 * Modified by :
 *    Lance E. Shepard     Rick Wise               Blake McBride
 *    CCI & RIT            CALCULON Corp.          3900 SW 52nd Ave. #803
 *    Rochester, NY        Rockville, MD 20850     Pembroke Park, FL 33023
 *
 *    John Stidd           Dale Loftis             Tom Hiscox
 *    1570 Neston Way      3020 Yorba Linda #F-16  136 Knorr Road
 *    Los Altos, CA 94022  Fullerton, CA 92631     Monroe, CT 06468  72106,3307
 *
 *    Monty J Clapham
 *    937 North Butrick
 *    Waukegan, Il 60085
 *
 * Special Notes:
 *    Compiles with the /AL large model switch to ensure enough
 *    space for the Table of Contents.  In MSC v5.1. malloc will
 *    default to _fmalloc or _nmalloc depending on compile line
 *    switch.
 *
 *    Tab stops on this version source code have been fully expanded
 *    to spaces This takes more space but eliminates the mess when you
 *    load it into your tab stops.
 *
 *************************************************************************/

/**************************************************************************
 *                         Program Revision History
 *
 *  By Dale Loftis
 *  10-3-87    Changed WIDTH to 132, Tabs are now every 4, modified so
 *             that temporary file is deleted if control - break is hit,
 *             Improved the clarity of the instructions if no parmeters
 *             are given. *
 *
 *  10-17-87   Changed so that no blank lines are printed in between
 *             structures.
 *
 *  11-06-87   Added -@ RESPONSE FILE feature since you couldn't fit all
 *             the filenames onto the command line, upto 400 files can
 *             be printed.
 *
 *             Also adding the "Reading file # x of y" string and made
 *             program check the amount of free disk space before writing
 *             the files as the program would write the output file and
 *             run out of disk space and just put alot of junk on the disk.
 *
 *  11-08-87   Ver 1.0  Added -h HEADER FILE, -* "Print All 'C' Files,
 *                      -x Extra Lines and Program Version Number
 *
 *
 *  01-10-88  Ver 1.1    Added call to setvbuf() to speed up file access.
 *
 *
 *  01-25-88  Ver 1.2    Added -t  for option of just table of contents list
 *
 *
 *  01-29-88  Ver 1.3    Fixed Bug.  funcname () would not be put in Table of
 *                         Contents because of space between funcname and ().
 *
 *
 *  04-26-88  Ver 1.4    Added -p* option for starting page number and changed
 *                       Table of Contents so File  Date and Time Shown.
 *
 *
 *  05-12-88  Ver 3.0    Added -q* option to print files by section number
 *                       rather than by page number.
 *                       Added -t- option to suppress the Table of Contents
 *
 *
 *  07-18-88  Ver 3.1    Fixed bug in printing by section where form feed was
 *                       not being output at the top of the file.
 *
 *
 *  (T. Hiscox)
 *  03-25-89  Ver 3.2    Translation to Microsoft C Version 5.1
 *                       Sort by actual function name(not case sensitive)
 *                       ignoring type cast, print aligned with function name.
 *                       {No messages on Warning Level 3 !!}
 *
 *
 *  (T. Hiscox)
 *  09-01-89  Ver 3.3    Add Cross Reference of function calls between modules
 *                       Printer Support for Epson and Hewlett Packard Deskjet
 *  (M. Clapham)
 *  05-19-90  Ver 3.4    Corrected Printer Codes/Dates/Pages from being printed
 *                       on all pages with the -c switch. Changed DumpTempFile()
 *                       to open in text mode for proper dump to be piped to
 *                       dos file (  > file.out ). Add Clear Screen routine
 *                       All code deleted, changed, or added, by me has been
 *                       commented out and left with my intials MJC for easy
 *                       identifing of changes.
 *
*************************************************************************/

/* --------------------------------------------------------------------- */
/*
 * Include Files for Library Function Prototypes 
 */
#include  <stdio.h>
#include  <stdlib.h>
#include  <fcntl.h>
#include  <malloc.h>
#include  <dos.h>
#include  <conio.h>
#include  <time.h>
#include  <string.h>
#include  <ctype.h>
#include  <io.h>

/* --------------------------------------------------------------------- */
/*
 * Some Valuable Definitions 
 */
#define      O_RAW  O_BINARY

#define      TEMPORARY_FILENAME   "CPR$.TMP"

#define      MAJOR_VERSION  3
#define      MINOR_VERSION  3

#define      MAX_S          256
#define      LENGTH         60
#define      WIDTH          132
#define      N_FILES        400
#define      TOC_LEN        1024
#define      MAX_FNAME_SIZE 50
#define      MAX_HDR_BYTES  500
#define      HEADER_SIZE    3

/* ------------------------------------------------------------------------- */
/*
 * These Macros are used to determine valid characters for Table of Contents
 */

/*
 * returns true if letter, digit, '_' , '*', or ' '
 */
#define isidchr(c)      (isalnum(c) || (c == '*') || (c == '_') || (c == ' '))

/*
 * returns true if letter, digit, '_', '*', '\t' or '\n'
 */
#define isiechr(c)      (isidchr(c) || c=='*' || c=='\t' || c=='\n')

/*
 * returns true is letter, digit, '_', '*'
 */
#define isnamechr(c)		(isalnum(c) || (c == '*') || (c == '_') )


/* --------------------------------------------------------------------- */
/*
 * Internal Program Function Prototypes 
 */
void main (int argc, char **argv);
void Usage (void);
void StartTempFile (void);
void EndTempFile (void);
void DumpTempFile (void);
int store_mod_call (struct cptr * modlist[], int m1, int m2);
int store_call (struct tnode * stree, int tindex);
int store_uses (struct tnode * stree, int tindex, int p);
int search_call (struct tnode * stree, int tindex, struct cptr ** cp);
int search_use (struct tnode * stree, int tindex, struct cptr ** cp);
void module_map (struct cptr * modlist[], char *title);
void proc_map (int (*sroutine) (),  char *title);
void DumpCrossReference (void);
void Done (void);
void List (void);
void NewPage (void);
void NewFile (void);
int PutLine (char *l);
void NewFunction (void);
void BreakPage (void);
void PutHeader (void);
int keycomp (char *elem1, char *elem2);
void write_xref (char *s);
int LooksLikeFunction (char *s);
void AddToTableOfContents (void);
void DumpTableOfContents (void);
char *strupper (char *str);
char *root_name (char *ns);
void SortTableOfContents (void);
char *EndComment (char *p, int *flag);
char *EndString (char *p, int *flag);
char *substr1 (char *string, int start, int end);
char *expand (char *string);
char *Rs (char *s);
int Cmemb (char a, char *b);
int StarDot (char *file_string);
void ReportMemoryError (void);
void ReadHeaderFile (char *header_file);
int ReadResponseFile (char *response_file);
int c_break (void);
void file_stats (char *filename);
int printer_code_init (void);
void normal_print (void);
void emphasized_print_on (void);
void emphasized_print_off (void);
void print_double_width (void);
void compressed_print (void);
void new_print_page (void);
void underline_on (void);
void underline_off (void);
void pspace (int ns);
char *strupper (char *str);
char *strsave (char *s);
void *strfree (char *s);
char *strcompress (char *s);
int CrossRefFunction (char *s);
int searchtree_index (struct tnode * stree, char *word, int tindex);
int searchtree_word (struct tnode * stree, char *word, int *tindex);
struct tnode *entertree (struct tnode * stree, char *word, int *lindex, int *tindex);
void clear_screen(void);
/* -------------------------------------------------------------------- */
/*
 * Global Variables 
 */
FILE *TempFile;			/* Program Temporary File       */
FILE *File;			/* Current File being worked on */
FILE *Rfile;			/* Response File handle         */
FILE *Hfile;			/* Header   File handle         */

struct find_t fblock;		/* struct used for calculating file size */
struct diskfree_t disk_space;	/* struct used for finding free disk space */

char file_time_buff[20];	/* file time is placed here      */
char file_date_buff[20];	/* file date is placed here      */

int Braces;			/* Keeps track of brace depth    */
int LineNumber;			/* Count output lines           */
int PageNumber = 1;		/* You figure this one out       */
int SectionNumber = 1;		/* and this one to               */
int PageLength = LENGTH;	/* Normal paper length           */
int PageWidth = WIDTH;		/* normal page width             */
int OnePerPage = 0;		/* non-zero puts 1 function/page */
int NumLines = 0;		/* non-zero puts line # in file  */
int Number = 1;
int WantSorted = 1;		/* Sort the table of contents   */
int PageEnd;			/* Accounts for space at bottom */
int InComment;			/* Indicates if inside Comment  */
int InXComment;			/* Indicates if inside Comment (Xref) */
int InString;			/* Indicates if inside String   */
int InXString;			/* Indicates in String for Xref Function */
int double_strike;		/* # of times + 1 to write filename             */
int header_flag;		/* non-zero enables print codes/Dates/Page #'s  */
int header_lines;		/* number of extra lines from header file       */
int extra_lines;		/* non-zero enables extra lines after functions */
int table_contents;		/* non-zero indicates Table of Contents Only    */
int sectionflg;                 /* non-zero indicates print by section          */

/**
 * MJC
 * int    nprinter = 1;
 */
int nprinter = 2;

/*
 * zero = Epson, 1 = Deskjet, 2 = ASCII(none)
 */

unsigned long memfree;		/* amount of local memory left for file buffer  */

char *TempName;			/* pointer to temporary filename   */
char *fname[N_FILES];		/* file names to be worked on      */
char *STDIN = "\n";		/* special string pointer value to */

/*
 * flag a file as being stdin
 */
char *Name;			/* Current file name               */
char *ProgName;			/* This programs Path and Name     */

char *Toc[TOC_LEN];		/* pointers to files in TOC        */
int TocPages[TOC_LEN];		/* page numbers                    */
int TocIndex[TOC_LEN];
int TocCount;			/* index into TOC arrays           */
char *cproc;
int nextf;			/* index into fname[]                */

char header_buffer[MAX_HDR_BYTES];	/* 500 byte buffer for user defined Page Headers */
char FunctionName[255];		/* Work area for parsing Func names */
char Todayv[45];		/* Today's Time and Date        */

/* --------------------------------------------------------------------- */
/**
 * The filenames[N_FILES][25] array allocates an array capable of holding
 * N_FILES filenames that are up to 25 characters long.
 *
 * Each filename in DOS can be 12 characters long and the response file
 * has a CRLF with each line and we also need a terminating zero
 * byte bringing the count up to 15. MAX_FNAME_SIZE bytes are provided
 * for a possible path
 */
/* --------------------------------------------------------------------- */

/*
 * char *filenames[N_FILES];
 */

/* --------------------------------------------------------------------- */
/**
 * The following string is the definition of tab stops.  Every 'T' is a
 * tab, any other character is not.  It is currently  defined as a tab
 * every 3 spaces.  The comment below helps if you want to modify this,
 * each row being 0+n, 50+n, 100+n, and 150+n, where n is the number
 * above the declaration.  Don't try to number each row with a comment,
 * because you'll notice that the '\'s make it one big string.
 */

char *TabDef = "\
--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T\
--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T--T";

/* ------------------------------------------------------------------------ */
#define  PRINTERS 3

struct esc_codes
{
    char *id;
    char *prt_init;
    char *prt_compress;
    char *prt_normal;
    char *prt_double_width_on;
    char *prt_double_width_off;
    char *prt_emphasize_on;
    char *prt_emphasize_off;
    char *prt_new_page;
    char *prt_underline_on;
    char *prt_underline_off;
};

struct esc_codes print_code[PRINTERS] =
{
      {  "Epson FX",
	 "\x01B\x040",
	 "\x00F",
	 "\x012",
	 "\x00E",
	 "\x014",
	 "\x01B\x045",
	 "\x01b\x046",
	 "\x00C",
	 "\x01B\x02D\x000",
	 "\x01B\x02D\x001"
      },
      {  "DeskJet ",
	 "\x01B(s16.67h0B\x01b&a10L",
	 "\x0",
	 "\x01B(s16.67h0B",
	 "\x01b(s5h3B",
	 "\x01B(s16.67h0B",
	 "\x01b(s3B",
	 "\x01b(s0B",
	 "\x01B&l0H",
	 "\x01b&d2D",
	 "\x01b&d@"
      },
      {  "ASCII Text",
	 "",
	 "",
	 "",
	 "",
	 "",
	 "",
	 "",
	 "\x0C",
	 "",
	 ""
      },
};

struct esc_codes *printer;

typedef char ckey_type[10];
ckey_type ckeys[] =
{
 "auto     ",
 "break    ",
 "case     ",
 "cdecl    ",
 "char     ",
 "const    ",
 "continue ",
 "default  ",
 "do       ",
 "double   ",
 "else     ",
 "enum     ",
 "extern   ",
 "far      ",
 "float    ",
 "for      ",
 "fortran  ",
 "goto     ",
 "huge     ",
 "if       ",
 "int      ",
 "interrupt",
 "long     ",
 "near     ",
 "pascal   ",
 "register ",
 "return   ",
 "short    ",
 "signed   ",
 "sizeof   ",
 "static	 ",
 "struct   ",
 "switch   ",
 "typedef  ",
 "union    ",
 "unsigned ",
 "void     ",
 "volatile ",
 "while    "};

#define  KEYSIZE  sizeof(ckeys) / sizeof(ckey_type)

/*------------------------------- tree types ------------------*/
struct cptr
{
    struct cptr *next_ptr;
    int proc;
};

struct entrytype
{
    char *proc;
    int module;
    unsigned int index;
    struct cptr *calls;
    struct cptr *uses;
};

struct tnode
{
    struct entrytype *this_entry;
    struct tnode *left;
    struct tnode *right;
};

struct tnode *proctree;
struct cptr *modcall[N_FILES];
struct cptr *moduse[N_FILES];

int module_index,
    proc_index,
    last_module_index,
    last_proc_index;

struct call_type
{
    int module;
    int proc;
    char *call;
};

#define  MAX_XREF 6144
struct call_type call_list[MAX_XREF];
int num_call;

/* ------------------------------------------------------------------------ */
/*****************************************************************************/
void main (argc, argv)
    int argc;
    char **argv;
{
    unsigned long int total_disk_free = 0;
    unsigned int ch;
    register int i;
    long thetime;

    nextf = 0;
    double_strike = 1;		/* default to no double strike       */
    header_flag = 1;		/* default to codes and headers      */
    header_lines = 0;		/* no extra lines from header file   */
    extra_lines = 0;		/* no extra lines after functions    */
    cproc = NULL;
    last_proc_index = last_module_index = 0;
    num_call = 0;
    TocCount = 0;
    proctree = NULL;
    memset (modcall, '\0', sizeof (modcall));
    memset (moduse, '\0', sizeof (moduse));

    /*
     * grab pointer to the programs name
     */
    ProgName = argv[0];

    /*
     * grab the DOS time and date
     */
    time (&thetime);

    /*
     * convert to ascii (LF at END OF STRING)
     */
    strcpy (Todayv, ctime (&thetime));

    /*
     * strip the annoying LF off
     */
    strtok (Todayv, "\n");

    /*
     * if only the program name given
     */
    if (argc == 1)
	/*
	 * if no args, show instructions
	 */
	Usage ();

    /**
     * MJC
     * fprintf (stderr, "\nCPR - A C Source Code Printing and Cross Reference Utility\nVersion 3.3  September 1989\n\n");
     * add clear_screen funciton
     */
    clear_screen();
    fprintf (stderr, "\n\nCPR - A  C  Source Code Printing and Cross Reference Utility\nVersion 3.4  May 1990\n\n");

    /*
     * other parse command line
     */
    for (i = 1; i < argc; i++)
    {
	/*
	 * check for switch flag "-"
	 */
	if (argv[i][0] == '-')
	{
	    /*
	     * and go thru the switches
	     */
	    switch (toupper (argv[i][1]))
	    {

	    /*
	     * Response file
	     */
	    case '@':
		nextf = ReadResponseFile (argv[i]);
		break;

	    /*
	     * "*.???" Print all specified files
	     */
	    case 'A':
		nextf = StarDot (argv[i]);
		break;

	    /*
	     * Print codes/Dates/Time and page #
	     */
	    case 'C':
		header_flag = 0;
		break;

	    /*
	     * Double Strike function names
	     */
	    case 'D':
		double_strike = 3;
		break;

	    /*
	     * Header file
	     */
	    case 'H':
		ReadHeaderFile (argv[i]);
		break;

	    /*
	     * Number of lines per page
	     */
	    case 'L':
		/*
		 * if -L ?? (ie.. a space)
		 */
		if (!argv[i][2])
		{
		    /*
		     * see if another argument
		     */
		    if (++i >= argc)
			/*
			 * if no then print Help
			 */
			Usage ();

		    /*
		     * is it a number?
		     */
		    if (!isdigit (argv[i][0]))
			/*
			 * No! Print Help
			 */
			Usage ();
		    /*
		     * assign number
		     */
		    PageLength = atoi (argv[i]);
		}  /* end of if */

		/*
		 * Switch came as -L?? form (with no space)
		 */
		else
		{
		    /*
		     * is it a number?
		     */
		    if (!isdigit (argv[i][2]))
			/*
			 * No! Print Help
			 */
			Usage ();
		    /*
		     * assign number
		     */
		    PageLength = atoi (&argv[i][2]);
		} /* end of else */
		break;

	    /*
	     * Number Source Lines
	     */
	    case 'N':
		NumLines = 1;
		break;

	    /*
	     * 1 function per page
	     */
	    case 'O':
		OnePerPage = 1;
		break;

	    /*
	     * Starting Page Number
	     */
	    case 'P':
		/*
		 * if -P ?? (ie.. a space)
		 */
		if (!argv[i][2])
		{
		    /*
		     * see if another argument
		     */
		    if (++i >= argc)
			/*
			 * if no then print Help
			 */
			Usage ();
		    /*
		     * is it a number?
		     */
		    if (!isdigit (argv[i][0]))
			/*
			 * No! Print Help
			 */
			Usage ();
		    /*
		     * assign number
		     */
		    PageNumber = atoi (argv[i]);
		} /* end of if */
		/*
		 * Switch came as -P?? form (with no space)
		 */
		else
		{
		    /*
		     * is it a number?
		     */
		    if (!isdigit (argv[i][2]))
			/*
			 * No! Print Help
			 */
			Usage ();
		    /*
		     * assign number
		     */
		    PageNumber = atoi (&argv[i][2]);
		} /* end of else */
		break;

	    /*
	     * Print by Section
	     */
	    case 'Q':
		/*
		 * if -Q ?? (ie.. a space)
		 */
		if (!argv[i][2])
		{

		    /*
		     * see if another argument
		     */
		    if (++i >= argc)
			/*
			 * if no then print Help
			 */
			Usage ();
		    /*
		     * is it a number?
		     */
		    if (!isdigit (argv[i][0]))
			/*
			 * No! Print Help
			 */
			Usage ();
		    /*
		     * assign number
		     */
		    SectionNumber = atoi (argv[i]);
		} /* end of if */
		/*
		 * Switch came as -Q?? form (with no space)
		 */
		else
		{
		    /*
		     * is it a number?
		     */
		    if (!isdigit (argv[i][2]))
			/*
			 * No! Print Help
			 */
			Usage ();
		    /*
		     * assign number
		     */
		    SectionNumber = atoi (&argv[i][2]);
		} /* end of else */
		sectionflg = 1;
		break;

	    /*
	     * Sorted Table of Contents
	     */
	    case 'S':
		WantSorted = 0;
		break;

	    /*
	     * Print Table of Contents Only
	     */
	    case 'T':
		/*
		 * enable flag
		 */
		table_contents = 1;
		/*
		 * if disabling TOC
		 */
		if (argv[i][2] == '-')
		    /*
		     * disable it
		     */
		    table_contents = -1;
		break;

	    /*
	     * Page Width
	     */
	    case 'W':
		/*
		 * if -W ?? (ie.. a space)
		 */
		if (!argv[i][2])
		{
		    /*
		     * see if another argument
		     */
		    if (++i >= argc)
			/*
			* if no then print Help
			*/
			Usage ();
		    /*
		     * is it a number?
		     */
		    if (!isdigit (argv[i][0]))
			/*
			 * No! Print Help
			 */
			Usage ();
		    /*
		     * assign number
		     */
		    PageWidth = atoi (argv[i]);
		} /* end of if */

		/*
		 * Switch came as -W?? form (with no space)
		 */
		else
		{
		    /*
		     * is it a number?
		     */
		    if (!isdigit (argv[i][2]))
			/*
			 * No! Print Help
			 */
			Usage ();
		    /*
		     * assign number
		     */
		    PageWidth = atoi (&argv[i][2]);
		}  /* end of else */
		break;

	    /*
	     * Extra Lines after Functions
	     */
	    case 'X':
		extra_lines = 1;
		break;

	    /*
	     * End of string?(out of enviroment)
	     */
	    case '\0':
		if (nextf >= N_FILES)
		{
		    fprintf (stderr, "%s: too many files\n", argv[0]);
		    exit (1);
		} /* end of if */

		/*
		 * any in excess goes to STDIN
		 */
		fname[nextf++] = STDIN;
		break;

	    /*
	     * No arguments given
	     */
	    default:
		/*
		 * show program instructions
		 */
		Usage ();
		break;
	    } /* end of switch */

	}  /* end of if */

	/*
	 * Argument (argv[i]) wasn't a switch parmeter
	 */
	else
	{
	    if (nextf >= N_FILES)
	    {
		fprintf (stderr, "%s: too many files\n", argv[0]);
		exit (1);
	    } /* end of if */
	    /*
	     * copy pointer to filename
	     */
	    fname[nextf++] = argv[i];
	} /* end of else */

    } /* end of for loop */

    /*
     * No files were specified
     */
    if (!nextf)
	/*
	 * so use STDIN as the input file
	 */
	fname[nextf++] = STDIN;

    PageEnd = PageLength - (1 + PageLength / 20);

    /*
     * open temporary Output file
     */
    StartTempFile ();

    printer_code_init ();

     /*
      * walk thru file list
      */
    for (i = 0; i < nextf; i++)
    {
	/*
	 * if no files
	 */
	if (fname[i] == STDIN)
	{
	    /*
	     * assign stdin as input
	     */
	    File = stdin;

	    /*
	     * assign name for printout
	     */
	    Name = "Standard Input";
	    fprintf (stderr, "Accepting Input from STDIN, ^Z to END\n");
	}  /* end of if */


	/*
	 * there was a filename
	 */
	else
	{
	    if ((File = fopen (Name = fname[i], "r")) == NULL)
	    {
		fprintf (stderr, "%s: Can't open file \"%s\"\n", ProgName, Name);
		/*
		 * keep going if not found
		 */
		continue;
	    } /* end of if */

	    /*
	     * file opened OK so read
	     */
	    else
	    {
		fprintf (stderr, "\nReading File #%3d of %3d: \" %s \"", i + 1, nextf, fname[i]);

		/*
		 * compute amount of free memory
		 */
		memfree = _memmax ();

		/*
		 * if to much memory then set max for setvbuf else take as much as u can
		 */
		memfree = ((memfree > 32767L) ? 32767L : memfree);

		if (setvbuf (File, NULL, _IOFBF, (int) memfree) != 0)
		{
		    fprintf (stderr, "Input Buffer failed to allocate.\n");
		} /* end of if */
	    }  /* end of else */

	} /* end of else */

	if (cproc)
	    free (cproc);
	cproc = NULL;
	module_index = i;
	/*
	 * Read File in and process data
	 */
	List ();
	/*
	 * if it's not STDIN
	 */
	if (File != stdin)
	{
	    /*
	     * close file & deallocate buffer
	     */
	    fclose (File);

	    /**
	     * MJC
	     * fprintf (stderr, " Table Entries: %4d  Cross Refs: %4d", TocCount, num_call);
	     */

	    fprintf (stderr, "\n\t\t\t\t\tTable Entries: %4d  Cross Refs: %4d", TocCount, num_call);
	} /* end of if */

    } /* end of for loop */

    if (PageNumber > 1 || LineNumber > 0)
	BreakPage ();

    /*
     * Close Temporary File
     */
    EndTempFile ();

    /**
     * MJC
     * fprintf (stderr, "\n\n");
     */
    fprintf (stderr, "\n");

    _dos_getdiskfree (0, &disk_space);

    /*
     * bytes per sector
     */
    total_disk_free = (((long) disk_space.bytes_per_sector *
			/*
			 * sectors per cluster
			 */
			(long) disk_space.sectors_per_cluster) *

			/*
			 * clusters available
			 */
			(long) disk_space.avail_clusters);

    _dos_findfirst (TEMPORARY_FILENAME, 0, &fblock);

    /*
     * see if copy can be output
     */
    if (fblock.size >= total_disk_free)
    {
	fprintf (stderr, "Insufficient Disk Space!\n\
                   \nMinimum of %lu bytes needed.\
                   \n%lu bytes are available\n\
                   \nDo You wish to continue? (Y/N)",
		 fblock.size, total_disk_free);

	do
	{
	    ch = toupper (getche ());
	} while ((ch != 'Y') && (ch != 'N'));

	if (ch != 'Y')
	    /*
	     * Delete Temporary File
	     */
	    Done ();

	fprintf (stderr, "\n\n");
    } /* end of if */

    /*
     * if TOC and file enabled or TOC only enabled
     */
    if ((!table_contents) || (table_contents == 1))
	/*
	 * Print Table of Contents
	 */
	DumpTableOfContents ();

    DumpCrossReference ();

    /*
     * if TOC and file enabled or no TOC
     */
    if ((!table_contents) || (table_contents == -1))
	/*
	 * then dump the file
	 */
	DumpTempFile ();

    /*
     * Clean things up and leave
     */
    Done ();

}  /* end of main () */

/*****************************************************************************/
/**
 * MJC
 * add clear_screen funciton below
 */
void Usage ()
{
    clear_screen();
    fprintf (stderr, "\nUsage: %s SWITCHES filename [...]", ProgName);
    fprintf (stderr, "\n----------------------------- VERSION %d.%d SWITCHES ------------------------\n", MAJOR_VERSION, MINOR_VERSION);
    fprintf (stderr, "           -@*    Replace * with Response Filename (1 filename per line    )\n");
    fprintf (stderr, "           -a*    Replace * with Drive Wildcard\n");
    fprintf (stderr, "           -c     Printer Codes/Dates/Page #'s off (default is Enabled     )\n");
    fprintf (stderr, "           -d     Double Strike Printing Enabled   (default is Disabled    )\n");
    fprintf (stderr, "           -h*    Replace * with Header Filename\n");
    fprintf (stderr, "           -l##   Replace ## with PageLength       (default = %3d          )\n", LENGTH);
    fprintf (stderr, "           -n     Numbers Source Lines             (Resets to 1 w/each file)\n");
    fprintf (stderr, "           -o     Places 1 function on a Page      (default is multiple    )\n");
    fprintf (stderr, "           -p*    Replace * with starting Page Number\n");
    fprintf (stderr, "           -q*    Replace * with starting Section Number\n");
    fprintf (stderr, "           -s     Sort Table of Contents by Page # (default is Func. Name  )\n");
    fprintf (stderr, "           -t     Print Table of Contents Only     (-t- suppress's Table   )\n");
    fprintf (stderr, "           -w##   Replace ## with Width            (default = %3d          )\n", WIDTH);
    fprintf (stderr, "           -x     Enables Extra Lines after Functions (default is Disabled )\n");
    fprintf (stderr, "          [...]   Means multiple filenames may be listed\n\n");
    fprintf (stderr, "               Response Files should have 1 filename per line up to %d      \n", MAX_FNAME_SIZE - 1);
    fprintf (stderr, "         characters followed by 1 CRLF and may contain up to %d filenames.  \n\n", N_FILES);
    fprintf (stderr, "               Header Files contain the Strings to Print at the top of      \n");
    fprintf (stderr, "         every page of the listing, entire file can be up to %d bytes.", MAX_HDR_BYTES);
    exit (1);
}

/*****************************************************************************/
void StartTempFile ()
{
	/*
	 * Temporary Filename
	 */
	TempName = TEMPORARY_FILENAME;

    if ((TempFile = fopen (TempName, "w")) == NULL)
    {
	fprintf (stderr, "%s: Can't open temp file!\n", ProgName);
	exit (1);
    } /* end of if */
}  /* end of StartTempFile () */

/*****************************************************************************/
void EndTempFile ()
{
        fclose (TempFile);
}  /* end of EndTempFile () */

/*****************************************************************************/
void DumpTempFile ()
{
    int fd,
        n;
    char buff[1025];
    /**
     * MJC
     * if ((fd = open (TempName, O_RDONLY | O_RAW)) == -1)
     */
    if ((fd = open (TempName, O_RDONLY | O_TEXT)) == -1)
    {
	fprintf (stderr, "%s: can't open temp file\n", ProgName);
	exit (1);
    }  /* end of if */

    fprintf (TempFile, "%s", printer->prt_init);
    while ((n = read (fd, buff, 1024)) != 0)
    {
	if (write (1, buff, n) == -1)
	{
	    fprintf (stderr, "%s: write error (1)\n", ProgName);
	    exit (1);
	}  /* end of if */
    }  /* end of while */

}  /* end of DumpTempFile () */

/*****************************************************************************/
/*
 * Enter module list with m1 calling or using m2
 */
int store_mod_call (modlist, m1, m2)
    struct cptr *modlist[];
int m1,
    m2;
{
    struct cptr *pc,
       *cp;

    if ((pc = (struct cptr *) malloc (sizeof (struct cptr))) == NULL)
	exit (0);
    pc->next_ptr = NULL;
    pc->proc = m2;
    if (modlist[m1])
    {
	cp = modlist[m1];
	do
	{
	    if (cp->proc == m2)
	    {
		free (pc);
		return (1);
	    } /* end of if */

	    if (cp->next_ptr)
		cp = cp->next_ptr;
	    else
		break;
	} while (cp);
	cp->next_ptr = pc;
    } /* end of if */
    else
	modlist[m1] = pc;
    return (1);

} /* end of store_mod_call () */

/*****************************************************************************/
/*
 * Enter tree with tindex, stores proc_index and module_index in Calls
 */
int store_call (stree, tindex)
    struct tnode *stree;
    int tindex;
{
    struct cptr *pc,
       *cp;

    if (stree)
    {
	if (stree->this_entry->index == tindex)
	{
	    if ((pc = (struct cptr *) malloc (sizeof (struct cptr))) == NULL)
		exit (0);
	    pc->next_ptr = NULL;
	    pc->proc = proc_index;
	    if (stree->this_entry->calls)
	    {
		cp = stree->this_entry->calls;
		do
		{
		    if (cp->proc == proc_index)
		    {
			free (pc);
			return (1);
		    }  /* end of if */

		    if (cp->next_ptr)
			cp = cp->next_ptr;
		    else
			break;
		} while (cp);
		cp->next_ptr = pc;
	    }  /* end of if */
	    else
		stree->this_entry->calls = pc;
	    return (1);

	} /* end of if */
	else
	{
	    if (store_call (stree->left, tindex) == 0)
		store_call (stree->right, tindex);
	}  /* end of else */
    } /* end of if */
    return (0);

}   /* end of store_call () */

/*****************************************************************************/
/*
 * Enter tree with tindex, stores proc_index and module_index in Uses area
 */
int store_uses (stree, tindex, p)
    struct tnode *stree;
    int tindex,
        p;
{
    struct cptr *pc,
       *cp;

    if (stree)
    {
	if (stree->this_entry->index == tindex)
	{
	    if ((pc = (struct cptr *) malloc (sizeof (struct cptr))) == NULL)
		exit (0);
	    pc->next_ptr = NULL;
	    pc->proc = p;
	    if (stree->this_entry->uses)
	    {
		cp = stree->this_entry->uses;
		do
		{
		    if (cp->proc == p)
		    {
			free (pc);
			return (1);
		    }  /* end of if */

		    if (cp->next_ptr)
			cp = cp->next_ptr;
		    else
			break;
		} while (cp);
		cp->next_ptr = pc;
	    } /* end of if */
	    else
		stree->this_entry->uses = pc;
	    return (1);
	} /* end of if */
	else
	{
	    if (store_uses (stree->left, tindex, p) == 0)
		store_uses (stree->right, tindex, p);
	}  /* end of else */
    } /* end of if */
    return (0);

}   /* end of store_uses () */

/*****************************************************************************/
/*
 * Enter tree with tindex, returns ptr to call structure
 */
int search_call (stree, tindex, pc)
    struct tnode *stree;
    int tindex;
    struct cptr **pc;
{
    if (stree)
    {
	if (stree->this_entry->index == tindex)
	{
	    *pc = stree->this_entry->calls;
	    return (1);
	} /* end of if */
	else
	{
	    if (search_call (stree->left, tindex, pc) == 0)
		search_call (stree->right, tindex, pc);
	} /* end of else */
    }
    return (0);

}  /* end of search_call () */

/*****************************************************************************/
/*
 * Enter tree with tindex, returns ptr to uses structure
 */
int search_use (stree, tindex, pc)
    struct tnode *stree;
    int tindex;
    struct cptr **pc;
{
    if (stree)
    {
	if (stree->this_entry->index == tindex)
	{
	    *pc = stree->this_entry->uses;
	    return (1);
	} /* end of if */
	else
	{
	    if (search_use (stree->left, tindex, pc) == 0)
		search_use (stree->right, tindex, pc);
	} /* end of if */

    } /* end of if */
    return (0);

}  /* end of search_use () */

/*****************************************************************************/
void module_map (modlist, title)
    struct cptr *modlist[];
char *title;
{
    register int i,
        j;
    struct cptr *pc,
       *cp;

    NewPage ();
    underline_on ();
    fprintf (TempFile, title);
    underline_off ();

    cp = modlist[0];
    for (i = 0; i < nextf; i++)
    {
	cp = modlist[i];
	if (cp)
	{
	    fprintf (TempFile, "          %-19.19s ", fname[i]);
	    j = 0;
	    pc = cp;
	    while (pc)
	    {
		if (j > 2)
		{
		    j = 0;
		    fprintf (TempFile, "\n                              ");
		    LineNumber++;
		    if (LineNumber + 5 > PageEnd)
		    {
			NewPage ();
			underline_on ();
			fprintf (TempFile, title);
			underline_off ();
			fprintf (TempFile, "\n                              ");
			LineNumber++;
			j = 0;
		    }
		}
		fprintf (TempFile, "%-19.19s ", fname[pc->proc]);
		j++;
		pc = pc->next_ptr;
	    }
	    /**
	     * MJC
	     * fprintf (TempFile, "\n\n");
	     * LineNumber++;
	     * LineNumber++;
	     */
	    fprintf (TempFile, "\n");
	    LineNumber++;
	}
    }
}

/*****************************************************************************/
void proc_map (sroutine, title )
int (*sroutine) ();
char *title;
{
    char ws[100];
    struct cptr *pc;
    register int i,
        j;
    int tproc;

    NewPage ();
    underline_on ();
    fprintf (TempFile, title);
    underline_off ();

    i = 0;
    do
    {
	tproc = TocIndex[i];
	if (tproc)
	{
	    searchtree_index (proctree, ws, tproc);
	    fprintf (TempFile, "          %-19.19s ", ws);
	    j = 0;

	    sroutine (proctree, tproc, &pc);
	    while (pc)
	    {
		searchtree_index (proctree, ws, pc->proc);
		if (j > 2)
		{
		    j = 0;
		    fprintf (TempFile, "\n                              ");
		    LineNumber++;
		    if (LineNumber + 5 > PageEnd)
		    {
			NewPage ();
			underline_on ();
			fprintf (TempFile, title);
			underline_off ();
			fprintf (TempFile, "\n                              ");
			LineNumber++;
			j = 0;
		    }
		}
		fprintf (TempFile, "%-19.19s ", ws);
		j++;
		pc = pc->next_ptr;
	    }
	    fprintf (TempFile, "\n");
	    LineNumber++;
	}
	else
	{
	    printf ("%s", Toc[i]);
	    LineNumber += 2;
	}
	if (LineNumber + 5 > PageEnd)
	{
	    NewPage ();
	    underline_on ();
	    fprintf (TempFile, title);
	    underline_off ();
	    j = 0;
	}
	i++;
    } while (i < TocCount);

}


/*****************************************************************************/
/*
 * strip out non-module calls (maybe use these later) build calls and
 * uses references
 */
void DumpCrossReference ()
{
    int i;

    PageNumber = -1;
    LineNumber = 0;
    /* default output to standard out     */
    TempFile = stdout;

    /**
     * MJC
     * fprintf (stderr, "\nLinking Cross References.....");
     */
    fprintf (stderr, "\n\n\tLinking Cross References.....\n\n");

    for (i = 0; i < num_call; i++)
    {
	proc_index = 0;
	searchtree_word (proctree, call_list[i].call, &proc_index);
	if (proc_index)
	{
	    store_call (proctree, call_list[i].proc);
	    store_uses (proctree, proc_index, call_list[i].proc);
	    store_mod_call (modcall, call_list[i].module, module_index);
	    store_mod_call (moduse, module_index, call_list[i].module);
	}
    }

    fprintf (TempFile, "%s", printer->prt_init);

   /*
    * complete the call list first, do the module to module maps
    */
    Name = "MODULE INTERACTION MAP";
    module_map (modcall, "          Module Name         Uses These Modules:\n\n");
    module_map (moduse, "          Module Name         Is Used by These Modules:\n\n");

   /*
    * dump procedure cross references will be in order to Table of Contents
    */

    Name = "PROCEDURE CROSS REFERENCE MAP";
    /**
     * MJC
     * proc_map (search_call, "          Procedure           Uses These Procedures:\n\n");
     *  proc_map (search_use, "          Procedure Name      Is Used By These Procedures:\n\n");
     */

    proc_map (search_call, "          Procedure           Uses These Procedures:\n");
    proc_map (search_use, "          Procedure Name      Is Used By These Procedures:\n");

    /*
     * make sure everything is written out
     */
    fflush (stdout);
    BreakPage ();

}

/*****************************************************************************/
void Done ()
{
    /*
     * if you created a file
     */
    if  (TempName)
	    /*
	     * Delete temporary file
	     */
	    unlink (TempName);
    /*
     * and leave the hard way
     */
    exit (0);
}

/*****************************************************************************/
/*
 * Process Incoming File
 */
void List ()
{
    register int bp;
    char buffer[257];
    char *p;


    /*
     * if by section
     */
    if (sectionflg)
    {
	/*
	 * Pages start at 1
	 */
	PageNumber = 1;
	/*
	 * Lines reset each time
	 */
	LineNumber = 0;
	/*
	 * form feed if allowed
	 */
	/**
	 * MJC
	 * MOVED to just below }
	 * BreakPage ();
	 */
    }

     BreakPage ();

    /*
     * bump page & output header
     */
    NewPage ();
    NewFile ();
    /*
     * no begin page and braces
     */
    bp = Braces = 0;

    /*
     * read a line upto 256 bytes
     */
    while (fgets (buffer, 256, File) != NULL)
    {
	/*
	 * Remove Trailing Spaces
	 */
	Rs (buffer);
	/*
	 * if beginning page
	 */
	if (bp)
	    /*
	     * setup for new function
	     */
	    NewFunction ();
	/*
	 * if going past page end
	 */
	if ((LineNumber + 1) > PageEnd)
	    /*
	     * start a new page
	     */
	    NewPage ();
	/*
	 * no braces and a function
	 */
	if (!Braces && LooksLikeFunction (buffer))
	{
	    /*
	     * then add to TOC
	     */
	    AddToTableOfContents ();
	}
	else
	    CrossRefFunction (buffer);

	bp = PutLine (buffer);
	LineNumber++;
    }
    /*
     * if by sections
     */
    if (sectionflg)
	/*
	 * Bump to next Section
	 */
	SectionNumber++;
}

/*****************************************************************************/
/*
 * Process New Page
 */
void NewPage ()
{
    if  (LineNumber > HEADER_SIZE)
    {
	if  (PageNumber >= 0)
	{
		/*
		 * bump Page Number
		 */
		PageNumber++;
	}
	/*
	 * form feed if allowed
	 */
	BreakPage ();
	/*
	 * set flag for a header
	 */
	LineNumber = 0;
    }
    /*
     * Each new page
     */
    if (LineNumber == 0)
	/*
	 * Outputs a Header
	 */
	PutHeader ();
}

/*****************************************************************************/
void NewFile ()
{
    register int i,
        l;
    char temp[80];

    if (TocCount >= TOC_LEN)
    {
	fprintf (stderr, "%s: too many table of contents entries\n", ProgName);
	exit (1);
    }

    /*
     * allocate some memory
     */
    if ((Toc[TocCount] = (char *) malloc (130)) == NULL)
	ReportMemoryError ();

    sprintf (Toc[TocCount], "\n\tModule: %s%s%s", printer->prt_double_width_on, Name, printer->prt_normal);

    /*
     * and findout how long it is
     */
    l = strlen (Name) * 3 + 6;
    /*
     * if not to long
     */
    if (l < 64)
    {
	/*
	 * calc. TAB increment
	 */
	i = (64 - l) / 8;
	/*
	 * keep within 64 columns
	 */
	for (l = 0; l < i; ++l)
	    /*
	     * and put some TAB's out
	     */
	    strcat (Toc[TocCount], "\t");
    }

    file_stats (Name);

    /**
     * MJC
     * ADD if (head_flag);
     */
    if (header_flag);
    {
      if (! sectionflg)
      {
	 /*
	  * Page #, Time & date
	  */
	 /**
	  * MJC
	  * sprintf (temp, "  Page %4d  (%s  %s)\n", PageNumber, file_date_buff, file_time_buff);
	  */
	 sprintf (temp, "\n\n" );
      }
      else
      {
	 /*
	  * Section, Time & Date
	  */
	 /**
	  * MJC
	  * sprintf (temp, "  Section %4d (%s  %s)\n", SectionNumber, file_date_buff, file_time_buff);
	  */
	 sprintf (temp, "\n\n" );
      }
    }

    /*
     * copy to main buffer
     */
    strcat (Toc[TocCount], temp);

    /*
     * bump the entries
     */
    ++TocCount;

    /*
     * if numbering lines
     */
    if (NumLines)
	/*
	 * reset linenumber
	 */
	Number = 1;

    /*
     * not in comment or string
     */
    InComment = InString = InXComment = InXString = 0;
}

/****************************************************************************/
int PutLine (l)
    register char *l;
{
    register char c;
    int bp;
    char *save;
    char *section,
       *p;
    int offset;
    char Digits[15];
    int Size;
    int pos;

    bp = 0;
    /*
     * bump thru the string
     */
    for (save = expand (l); ((c = *l) != 0); ++l)
    {
	/*
	 * if your in a comment
	 */
	if (InComment)
	    /*
	     * process till your out
	     */
	    l = EndComment (l, &InComment);

	else
	{
	    /*
	     * if your in a string
	     */
	    if (InString)
	       /*
		* process till your out
		*/
	       l = EndString (l, &InString);

	       /*
		* it must be something
		*/
	    else
	    {
	       switch (c)
	       {
		  /*
		   * curly brace IN
		   */
		  case '{':
		     /*
		      * increment brace count
		      */
		     ++Braces;
		     break;

		  /*
		   * curly brace OUT
		   */
		  case '}':
		     /*
		      * coming out yet?
		      */
		     if (--Braces == 0)
		     {
			/*
			 * is it structure?
			 */
			if (*(l + 1) == ';')
			   /*
			    * continue thru string
			    */
			   ++l;

			else
			   /*
			    * if option enabled
			    */
			   if (extra_lines)
			      /*
			       * flag for extra lines
			       */
			      bp = 1;
		     } /* end of if */
		     break;

		  /*
		   * Forward Slash
		   */
		  case '\'':
		     /*
		      * continue thru string
		      */
		     ++l;
		     break;

		  /*
		   * Quotation Mark
		   */
		  case '"':
		     /*
		      * must be in a string
		      */
		     InString = 1;
		     break;

		  /*
		   * Start of Comment
		   */
		  case '/':
		     /*
		      * is next character * ?
		      */
		     if (*(l + 1) == '*')
		     {
			/*
			 * must be in a comment
			 */
			InComment = 1;
			/*
			 * continue thru string
			 */
			++l;
		     } /* end of if */
		     break;

	       } /* end of switch */


	    } /* end of else */

	 } /* end of else */

    } /* end of for loop */

    /*
     * if line numbering enabled
     */
    if (NumLines)
    {
	/*
	 * make a string
	 */
	sprintf (Digits, "[%4d]  ", Number);
	/*
	 * calc. its size
	 */
	Size = strlen (Digits);
    } /* end of if */

    else
	/*
	 * else it's size is 0
	 */
	Size = 0;

    /*
     * if original strings size and the size of numbers are
     * bigger than the width
     */
    if (strlen (save) + Size > PageWidth)
    {
	section = substr1 (save, 0, PageWidth - Size);

	if (section[strlen (section) - 1] != ' ')

	    if (NULL == (p = strrchr (section, ' ')))
		offset = strlen (section);

	    else
		offset = p - section;

	else
	    offset = strlen (section) - 1;

	section[offset] = NULL;

	/*
	 * if line numbering is enabled
	 */
	if (NumLines)
	    fprintf (TempFile, "[%4d]  %s\n", Number++, section);

	else
	    fprintf (TempFile, "%s\n", section);

	pos = offset + 1;

	do
	{
	    section = substr1 (save, pos, pos + PageWidth - 8);
	    if (strlen (save) - pos + 8 > PageWidth)
		if (section[strlen (section) - 1] != ' ')
		    if (NULL == (p = strrchr (section, ' ')))
			offset = strlen (section);
		    else
			offset = p - section;
		else
		    offset = strlen (section) - 1;
	    else
		offset = strlen (section);

	    section[offset] = NULL;

	/** ----------------------------------------------------------------- **/

	/*
	 * Removed this for some reason
	/*
	/*           if (section[strlen(section) - 1] == '\n')
	 *                 section[strlen(section) - 1] = NULL;
	 */

	/*
	 * Line is to long for pagewidth so continue on next line
	 */
	    fprintf (TempFile, "C       %s\n", section);
	    if (++LineNumber > PageEnd)
		NewPage ();

	} while ((pos += offset + 1) < strlen (save));

    } /* end of if */

    else
    {
	/*
	 * if line numbering enabled
	 */
	if (NumLines)
	    fprintf (TempFile, "[%4d]  %s\n", Number++, save);
	else
	    fprintf (TempFile, "%s\n", save);

    } /* end of else */

    return (bp);
}

/*****************************************************************************/
/*
 * Process New Function
 */
void NewFunction ()
{
    register int i;
    /*
     * 49.5 lines
     */
    if (LineNumber > (PageLength * 3 / 4))
	NewPage ();
    else
    {
	/*
	 * if mult. functions per page
	 */
	if (!OnePerPage)
	{
	    for (i = 0; i < (PageLength / 7); ++i)
		/*
		 * add extra lines
		 */
		putc ('\n', TempFile);

	    /*
	     * set line counter also
	     */
	    LineNumber += PageLength / 7;
	}
	/*
	 * otherwise its 1 func. per page
	 */
	else
	    NewPage ();
    }
}

/*****************************************************************************/
/*
 * Process Page Break
 */
void BreakPage ()
{
    /*
     * if its allowed send a FORM FEED
     */
    if  (header_flag)
	    new_print_page ();
}

/*****************************************************************************/
void PutHeader ()
{
    register int i,
        l;

    putc ('\n', TempFile);

    if (header_flag)
    {
	print_double_width ();
	l = strlen (Name) * 3;

	/*
	 * Write out Filename or TOC
	 */
	fprintf (TempFile, "%s", Name);
	normal_print ();
    }
    /*
     * if you're not on the TOC pages
     */
    if (PageNumber > 0)
    {
	for (i = (l + 7) / 8; i < 6; ++i)
	    /*
	     * Tab out for position
	     */
	    putc ('\t', TempFile);

       /**
	* MJC
	* ADD if (head_flag);
	*/
	if (header_flag)
	{
	    if (! sectionflg)
	    {
	       fprintf (TempFile, "%s\tPage: %d\n\n\n", Todayv, PageNumber);
	    }
	    else
	    {
	       fprintf (TempFile, "%s\tSection %3d - %d\n\n\n", Todayv, SectionNumber, PageNumber);
	    }
	 }
	/*
	 * Header for TOP OF PAGE
	 */
	if (header_lines)
	    fprintf (TempFile, "%s", header_buffer);
    }
    /*
     * 1st line of Table of Contents page gets the time and date
     */
    else
    {
	/*
	 * Tab out for time/date
	 */
	for (i = (l + 7) / 8; i < 6; ++i)
	    putc ('\t', TempFile);

	/*
	 *Todays Time and Date
	 */
	/**
	 * MJC
	 * ADD if(heder_flag)
	 * ADD fprintf (TempFile, "\n\n");
	 * fprintf (TempFile, "\n%s\n", Todayv);
	 */
	if(header_flag)
	   fprintf (TempFile, "\t\t\t\t%s\n\n", Todayv);
	else
	   fprintf (TempFile, "\n\n");

	/*
	 * Header for TOP OF PAGE
	 */
	if (header_lines)
	    fprintf (TempFile, "%s", header_buffer);
    }
    /*
     * bump line number counter
     */
    LineNumber += (HEADER_SIZE + header_lines);
}

/*****************************************************************************/
int keycomp (elem1, elem2)
    char *elem1,
       *elem2;
{
    int i;

    i = strcmp (strcompress (elem1), strcompress (elem2));
    return (i);
}

/*****************************************************************************/
void write_xref (s)
    char *s;
{
    int i;

    if (cproc)
    {
	if (strlen (s = strcompress (s) ) )
	{
	    if (bsearch (s, ckeys, KEYSIZE, sizeof (ckey_type), keycomp) == NULL)
	    {
	    /*
	     * printf(xref,"%3d %3d %-20.20s\n",module_index,proc_index, s );
	     */
		if (num_call < MAX_XREF - 1)
		{
		    call_list[num_call].module = module_index;
		    call_list[num_call].proc = proc_index;
		    call_list[num_call].call = strsave (s);
		    num_call++;
		}
	    }
	}
    }
}

/*****************************************************************************/
LooksLikeFunction (s)
    char *s;
{
    char *p;
    char *save;
    int nosl,
        nolp,
        norp,
        flg;
    int AddOne = 0;
    /*
     * if you're in a comment or in a string
     */
    if (InComment || InString)
	/*
	 * just leave and return 0
	 */
	return (0);
    /*
     * assign pointer to function name string
     */
    p = FunctionName;
    /*
     * save address of string
     */
    save = s;
    /*
     * no left or right paren or no slash
     */
    nosl = nolp = norp = 0;

    flg = 1;
    /*
     * go until end of string or flag is set
     */
    for (; *s && flg; s++)
    {
	/*
	* switch on each character in string
	*/
	switch (*s)
	{
	/*
	 * is it a left paren?
	 */
	case '(':
	    /*
	     * if no left paren already
	     */
	    if (!nolp)
	    {
		/*
		 * flag that you've got one
		 */
		nolp = 1;

		/*
		 * 3/4/87 added following line to include ARGS in TOC
		 */
		/*
		 * move byte of ARG into function buffer
		 */
		*p++ = *s;
	    }
	    else
		/*
		 * or return if you've already had one
		 */
		return (0);
	    break;
	/*
	 * is it a right paren?
	 */
	case ')':
	    /*
	     * if already a left paren and no right yet
	     */
	    if (nolp && !norp)
	    {
		/*
		 * flag that you've got a right paren
		 */
		norp = 1;

		/*
		 * 3/4/87 added following line to include ARGS in TOC
		 */
		/*
		 * move byte of ARG into function buffer
		 */
		*p++ = *s;
	    }
	    else
		/*
		 * or return if conditions were'nt right
		 */
		return (0);
	    break;
	/*
	 * everything comes thru here
	 */
	default:
	    /*
	     * if no left paren yet
	     */
	    if (!nolp)
	    {
		/*
		 * is it alpha,digit,'_*\t\n'?
		 */
		if (isiechr (*s))
		{
		    /*
		     * is it alpha,digit,'_' or ' '?
		     */
		    if (isidchr (*s))
			/*
			 * if it is store the byte for TOC
			 */
			*p++ = *s;
		    else
			/*
			 * start over if '*\t\n' & nolp
			 */
			p = FunctionName;
		    break;
		}
		else
		    /*
		     * return if conditions weren't right
		     */
		    return (0);
	    }
	   /*
	    * if no right paren yet
	    */
	    if (!norp)
	    {
		/*
		 * is it alpha,digit,'_ *\t\n'?
		 */
		if (isiechr (*s) || *s == ',')
		{
		    /*
		     * 3/4/87 added following line to include ARGS in TOC
		     */
		    /*
		     * move byte of ARG in function buffer
		     */
		    *p++ = *s;
		    break;
		}
		else
		    /*
		     * return if conditions weren't right
		     */
		    return (0);
	    }

	    /*
	     * is it ' \t\n\r' or no slash yet
	     */
	     if (Cmemb (*s, " \t\n\r") && !nosl)

	    /*
	     * is it a '/' and no slash yet
	     */
	    if (*s == '/' && !nosl)
	    {
		/*
		 * flag that you've got one
		 */
		nosl = 1;
		break;
	    }

	    /*
	     * if its '*' & you got a slash already
	     */
	    if (*s == '*' && nosl)
	    {
		/*
		 * set flag to abort loop
		 */
		flg = 0;
		break;
	    }
	    /*
	     * return if conditions not right
	     */
	    return (0);

	}   /* end of switch */
    }    /* end of for loop */

    /*
     * return if no left paren found
     */
    if (nolp != 1)
	return (0);
    /*
     * else terminate function buffer
     */
    *p = '\0';

    /*
     * if line numbering enabled
     */
    if (NumLines)
    {

    /*
     * the following code indents the Funtion Names with Line Numbering
     * It is removed in favor of inserting a blank line
     */
    /***
       *   sprintf (Digits,"[%4d]  ", Number);
       *   Cnt = strlen(Digits) + AddOne;
       *   while (Cnt-- > 0)
       *       putc (' ', TempFile);
       *   AddOne = 0;
      ***/
	putc ('\n', TempFile);
	LineNumber++;
    }

   /*
    * This will cause the function name part of the line to be double striken.
    */
    if (double_strike == 3)
    {
	/*
	 * double strike up to func name
	 */
	while (*save && *save != '(')
	    putc (*save++, TempFile);
	/*
	 * use CR for doublestrike
	 */
	putc ('\r', TempFile);
    }
    return (1);
}

/****************************************************************************/
/*
 * Enter tree with tindex, returns with alfa string corresponding
 */
int searchtree_index (stree, word, tindex)
    struct tnode *stree;
    char *word;
    int tindex;
{
    if (stree)
    {
	if (stree->this_entry->index == tindex)
	{
	    strcpy (word, stree->this_entry->proc);
	    module_index = stree->this_entry->module;
	    return (1);
	}
	else
	{
	    if (searchtree_index (stree->left, word, tindex) == 0)
		searchtree_index (stree->right, word, tindex);
	}
    }
    return (0);
}

/*****************************************************************************/
/*
 * enter with word, returns index
 */
int searchtree_word (stree, word, tindex)
    struct tnode *stree;
    char *word;
    int *tindex;
{
    if (stree)
    {
	if (stricmp (stree->this_entry->proc, word) == 0)
	{
	    proc_index = stree->this_entry->index;
	    module_index = stree->this_entry->module;
	    *tindex = stree->this_entry->index;
	    return (1);
	}
	else
	{
	    if (searchtree_word (stree->left, word, tindex) == 0)
		searchtree_word (stree->right, word, tindex);
	}
    }
    return (0);
}

/*****************************************************************************/
struct tnode *entertree (stree, word, lindex, tindex)
    struct tnode *stree;
    char *word;
    int *lindex,
       *tindex;
{
    if (stree == NULL)
    {
	stree = (struct tnode *) malloc (sizeof (struct tnode));
	stree->this_entry = (struct entrytype *) malloc (sizeof (struct entrytype));
	stree->left = NULL;
	stree->right = NULL;
	stree->this_entry->calls = NULL;
	stree->this_entry->uses = NULL;
	stree->this_entry->proc = strsave (word);
	stree->this_entry->module = module_index;
	(*lindex)++;
	stree->this_entry->index = *lindex;
	*tindex = *lindex;
    }
    else
	switch (strcmp (stree->this_entry->proc, word))
	{
	case 0:
	    *tindex = stree->this_entry->index;
	    break;
	case 1:
	    stree->left = entertree (stree->left, word, lindex, tindex);
	    break;
	default:
	    stree->right = entertree (stree->right, word, lindex, tindex);
	}
    return (stree);
}

/*****************************************************************************/
void AddToTableOfContents ()
{
    register int l;
    register char *p;

    if (TocCount >= TOC_LEN)
    {
	fprintf (stderr, "%s: too many table of contents entries\n", ProgName);
	exit (1);
    }
    l = strlen (FunctionName);
    p = Toc[TocCount] = (char *) malloc (l + 1);
    if (p == NULL)
	ReportMemoryError ();
    strcpy (p, FunctionName);
    TocPages[TocCount] = PageNumber;

   /*
    * build a tree of procedure names and modules
    */
    if (cproc)
	free (cproc);
    cproc = strsave (root_name (FunctionName));
    if (p = strchr (cproc, '('))
    {
	*p = ' ';
	while (*p && *p != ')')
	    *p++ = ' ';
	if (*p == ')')
	    *p = ' ';
    }
    cproc = strcompress (cproc);
    proctree = entertree (proctree, cproc, &last_proc_index, &proc_index);
    TocIndex[TocCount] = proc_index;
    TocCount++;
}

/****************************************************************************/
void DumpTableOfContents ()
{
    register int i,
        j,
        l;
    char *rn;

    /*
     * if there nothing to print
     */
    if (TocCount == 0)
	/*
	 * then just return
	 */
	return;

    /*
     * if you wanted it sorted by page number
     */
    if (WantSorted)
	/*
	 * then we must sort it first
	 */
	SortTableOfContents ();
    /*
     * give me a title for the page
     */
    Name = "TABLE of CONTENTS";

    /*
     * Table of Contents pages are not numbered
     */
    PageNumber = -1;
    /*
     * and neither are the lines
     */
    LineNumber = 0;
    /*
     * default output to standard out
     */
    TempFile = stdout;

    fprintf (TempFile, "%s", printer->prt_init);
    /*
     * start out creating a new page
     */
    NewPage ();

    for (i = 0; i < TocCount; ++i)
    {
	if (Toc[i][0] == '\n')
	{
	    if ((LineNumber + 5) > PageEnd)
		NewPage ();
	    printf ("%s", Toc[i]);
	    LineNumber += 2;
	    continue;
	}

	/*
	 * if going off end of page
	 */
	if (++LineNumber > PageEnd)
	    /*
	     * start a new page
	     */
	    NewPage ();
	rn = root_name (Toc[i]);
	l = ((rn - Toc[i]) < 20) ? 20 - (rn - Toc[i]) : 0;
	for (j = 0; j < l; j++)
	    putchar (' ');
	printf ("%s ", Toc[i]);

	/*
	 * length if function name with parms
	 */
	l += strlen (Toc[i]);

	/*
	 * put dots out to column 68
	 */
	for (j = l; j < 68; ++j)
	    /*
	     * put dots out to page num. in TOC
	     */
	    putchar ('.');

	/*
	 * print page number for function
	 */
	printf (" %d\n", TocPages[i]);
    }

    /*
     * make sure everything is written out
     */
    fflush (stdout);
    BreakPage ();
}

/*****************************************************************************/
/*********************************************************/
/* Search the Table of Contents Entry and return a    */
/* pointer to the root name part of the entry (v3.2)  */
/*********************************************************/
char *root_name (ns)
    char *ns;
{
    char *p;

    p = ns;
    do
    {
	p++;
    } while ((*p != '(') && (*p));
    if (*p)
    {
	do
	{
	    p--;
	} while ((!isnamechr (*p)) && (p > ns));
	do
	{
	    p--;
	} while ((*p != ' ') && (p > ns));
    }
    if (*p == ' ')
	p++;
    if (*p == '*')
	p++;
    return (p);
}

/****************************************************************************/
/*
 * Sort Table of Contents by Funtion Name 
 */
void SortTableOfContents ()
{
    register int i,
        tempint;
    char *tempchar;
    int flag;

    /**
     * MJC
     * fprintf (stderr, "\nSorting Table of Contents.....");
     */
    fprintf (stderr, "\n\n\tSorting Table of Contents.....");

    do
    {
	/*
	 * default to already sorted
	 */
	flag = 0;

	/*
	 * look at them all
	 */
	for (i = 0; i < TocCount - 1; i++)
	{
	    if (Toc[i][0] == '\n' || Toc[i + 1][0] == '\n')
		/*
		 * don't sort across file names
		 */
		continue;
	    /*
	     * compare the strings
	     */
	    if (stricmp (root_name (Toc[i]), root_name (Toc[i + 1])) > 0)
	    {
		/*
		 * copy to temp pointer
		 */
		tempchar = Toc[i];
		/*
		 * swap the pointers in array
		 */
		Toc[i] = Toc[i + 1];
		/*
		 * put the temp pointer back
		 */
		Toc[i + 1] = tempchar;
		/*
		 * grab the page number
		 */
		tempint = TocPages[i];
		/*
		 * swap page numbers
		 */
		TocPages[i] = TocPages[i + 1];
		/*
		 * put the temp page number back
		 */
		TocPages[i + 1] = tempint;
		tempint = TocIndex[i];
		TocIndex[i] = TocIndex[i + 1];
		TocIndex[i + 1] = tempint;
		/*
		 * indicate you've swapped
		 */
		flag = 1;
	    } /* end of if */

	} /* end of for loop */

    } while (flag);               /* go until no more swaps can be made */
}

/****************************************************************************/
/*
 * Process string until you come out of COMMENT
 */
char *EndComment (p, flag)
    char *p;
    int *flag;
{
    char c;
    /*
     * while there are chars to look at
     */
    while ((c = *p++))
	/*
	 * and there a splat or a slash
	 */
	if (c == '*' && *p == '/')
	{
	    /*
	     * say your not in the comment
	     */
	    *flag = 0;
	    /*
	     * and leave the loop
	     */
	    break;
	}
    /*
     * returning the new pointer addr.
     */
    return (p - 1);
}

/*****************************************************************************/
/*
 * Process string until you come out of STRING
 */
char *EndString (p, flag)
    char *p;
    int *flag;
{
    register char c;

    /*
     * while there are chars to look at
     */
    while ((c = *p++) != 0)
	/*
	 * and forward slashs for next line
	 */
	if (c == '\\')
	{
	    /*
	     * means "Just read on McDuck"
	     */
	    continue;
	}
	else
	/*
	 * if you found the ending quote
	 */
	if (c == '"')
	{
	    /*
	     * say your not in a string anymore
	     */
	     *flag = 0;

	    /*
	     * and leave the loop
	     */
	    break;
	}
    /*
     * returning the new pointer addr.
     */
    return (p - 1);
}

/****************************************************************************/
/*
 * This is the function substr1().  The calling sequence is: 
 *
 * substr1(string, startpos, endpos). 
 *
 * The function returns a pointer to a static string (written over - on subsequent calls) which is a substring of the string `string'
 * starting at `startpos' (the first position is 0 (zero)) and ending at `endpos' (non-inclusive).  All arguments must be present
 * or strange things happen with the system stack. 
 *
 * An example of the use is: 
 *
 * x = substr1(string, 2, 5); (where string == "This is a test.") 
 *
 * This call returns a pointer to: "is " An error code of -1 is returned is the `endpos' is greater than `startpos' 
 *
 * Lance E. Shepard 
 */
char *substr1 (string, start, end)
    char *string;
    int start;
    int end;
{
    static char retstr[MAX_S];
    int loop1;
    int loop2;

    if (end < start)
    {
	exit (-1);
    }

    for (loop2 = 0; loop2 < MAX_S; loop2++)
	retstr[loop2] = NULL;

    for (loop1 = start, loop2 = 0; string[loop1] != NULL &&
	 loop1 < end && loop2 <= MAX_S; loop1++, loop2++)
	retstr[loop2] = string[loop1];

    retstr[++loop2] = NULL;
    return (retstr);
}

/****************************************************************************/
/*
 * This is the function `char *expand().'  This function takes as an argument
 * a NULL terminated string and replaces all occurances
 * of the tab character with 8 (eight) spaces.  The function returns a pointer
 * to a static string which is overwritten on
 * subsequent calls. 
 */
/****************************************************************************/
char *expand (string)
    char *string;
{
    int count;
    static char retstr[MAX_S];

    for (count = 0; count < MAX_S; retstr[count++] = NULL);

    for (count = 0; *string != NULL; count++, string++)
    {
	if (*string == '\t')
	{
	    retstr[count] = ' ';

	    /*
	     * while (((count + 1) % 8) != 0)
	     */
	    while (TabDef[count] != 'T')
		retstr[++count] = ' ';
	}
	else
	    retstr[count] = *string;
    }
    retstr[count] = NULL;
    return (retstr);
}

/****************************************************************************/
/*
 * strip trailing blanks from string
 */
char *Rs (s)
    char s[];

{
    int n;

     /**
      * MJC ADDED
      */
      for (n = strlen (s) - 1; n >= 0 && iscntrl(s[n]); n--)
	if(s[n]== '\n')
		s[n] = '\0';

    /*
     * find the last space in the string
     */
    for (n = strlen (s) - 1; n >= 0 && isspace (s[n]); n--)
	/**
	 * MJC
	 * ;
	 */
    /*
     * plop a null char on top of it
     */
    s[n + 1] = '\0';

    /*
     * return pointer to string
     */
    return (s);
}

/****************************************************************************/
/*
 * is character "a" a member of string "b"
 */
int Cmemb (a, b)
    char a,
       *b;
{
    /*
     * go until the null character
     */
    while (*b)
	/*
	 * looking for the character and bumping the pointer
	 */
	if (a == *b++)
	    /*
	     * returning 1 if its found
	     */
	    return (1);
    /*
     * return 0 if "a" was a non-zero character
     */
    return ((!a) ? 1 : 0);
}

/****************************************************************************/
StarDot (file_string)
    char *file_string;
{
    unsigned long int total_file_size = 0;
    unsigned long int total_disk_free = 0;
    int done;
    int i = 0;

    /*
     * bump past -A switch
     */
    file_string += 2;

    /*
     * Attempt 1st file read
     */
    if ((done = _dos_findfirst (file_string, 0, &fblock)) != 0)
    {
	fprintf (stderr, "No Files Found\n");
	exit (1);
    }
    /*
     * go until done or to many files
     */
    while (!done && i < N_FILES)
    {
      /*
       * filenames[i] = strsave(fblock.name);
       */
      /*
       * strcpy(&filenames[i][0],fblock.name);
       */
       /*
	* copy names into array
	*/
	fname[i] = strsave (fblock.name);
	i++;
	/*
	 * keep running total of filesizes
	 */
	total_file_size += fblock.size;
	/*
	 * read next availables file info
	 */
	done = _dos_findnext (&fblock);
    }
    if (i == N_FILES)
	fprintf (stderr, "Printing ONLY 1st %d Files\n", N_FILES);

     /*
      * find free disk space
      */
    _dos_getdiskfree (0, &disk_space);

    /*
     * convert to bytes
     */
    total_disk_free = (((long) disk_space.bytes_per_sector *
			(long) disk_space.sectors_per_cluster) *
		       (long) disk_space.avail_clusters);

    if ((total_file_size * 2) >= total_disk_free)
    {
	fprintf (stderr, "Insufficient Disk Space!\nMinimum of %lu bytes needed.\n\nYou Have %lu bytes free\n", total_file_size * 2, total_disk_free);
	exit (1);
    }
    /*
     * return new number of files
     */
    return (i);
}

/*****************************************************************************/
void ReportMemoryError ()
{
     /**
      * MJC
      * fprintf (stderr, "Memory Allocation Error\n");
      */
      fprintf (stderr, "\nMemory Allocation Error\n");

     /*
      * if there is a file to work on
      */
    if (TempName)
    {
	/*
	 * Make sure file is closed
	 */
	EndTempFile ();
	/*
	 * Delete temporary file
	 */
	unlink (TempName);
    }
    exit (1);
}

/*****************************************************************************/
void ReadHeaderFile (header_file)
    char *header_file;
{
    extern char header_buffer[MAX_HDR_BYTES];
    extern int header_lines;
    int ch;
    int i = 0;
    int x = 0;
    char *string;

    /*
     * strip the "-h" from the string
     */
    header_file += 2;
    if ((Hfile = fopen (header_file, "r")) == NULL)
    {
	fprintf (stderr, "Can't Open Header File %s\n", header_file);
	exit (1);
    }
    else
    {
    /*
     * read entire file that is up to MAX_HDR_BYTES characters long
     */
	while ((ch = fgetc (Hfile)) != EOF)
	{
	    header_buffer[i++] = (unsigned char) ch;
	    if (i < MAX_HDR_BYTES)
		continue;
	    else
		break;
	}
	/*
	 * terminate buffer
	 */
	header_buffer[i] = '\0';

	/*
	 * and close the file
	 */
	fclose (Hfile);

	for (x = 0, string = header_buffer; x <= i; x++)
	{
	    /*
	     * count number of line feeds
	     */
	    if (*string++ == '\n')
		/*
		 * tell the NewPage Function
		 */
		header_lines++;
	}
    }
}

/*****************************************************************************/
ReadResponseFile (response_file)
    char *response_file;
{
    /*
     * area for string to be read into
     */
    char fpntr[80];

    unsigned long int total_file_size = 0;
    unsigned long int total_disk_free = 0;

    int i = 0;

    /*
     * strip the "-@" from the string
     */
    response_file += 2;
    if ((Rfile = fopen (response_file, "r")) == NULL)
    {
	fprintf (stderr, "Can't Open Response File %s\n", response_file);
	/*
	 * terminate abruptly
	 */
	exit (1);
    }
    else
    {
      /*
       * read a entire line from the Rfile that is up to 80 characters long
       */
	while (fgets (fpntr, 80, Rfile) != NULL && i < N_FILES)
	{
	    /*
	     * strtok(fpntr," \n");/
	     */
	    /*
	     * strip to LF or space
	     */

	    fname[i] = strsave (strtok (fpntr, " \n"));

	    /*
	     * strcpy(&filenames[i][0],fpntr);
	     */
	    /*
	     * copy names into array
	     */
	    /*
	     * fname[i] = filenames[i];
	     */
	    /*
	     * copy pointer to name
	     */

	    i++;

	    /*
	     * read file size in
	     */
	    _dos_findfirst (fpntr, 0, &fblock);

	    /*
	     * keep running total of filesizes
	     */
	    total_file_size += fblock.size;
	}

	/*
	 * and close the file
	 */
	fclose (Rfile);

	/*
	 * find free disk space
	 */
	_dos_getdiskfree (0, &disk_space);

	/*
	 * convert to bytes
	 */
	total_disk_free = (((long) disk_space.bytes_per_sector *
			    (long) disk_space.sectors_per_cluster) *
			   (long) disk_space.avail_clusters);

	if ((total_file_size * 2) >= total_disk_free)
	{
	    fprintf (stderr, "Insufficient Disk Space!\nMinimum of %lu bytes needed.\
                      \nYou Have %lu bytes free\n",
		     total_file_size * 2, total_disk_free);
	    exit (1);
	}
	/*
	 * return new number of files
	 */
	return (i);
    }
    return (0);
}

/*****************************************************************************
    This routine is executed when the CONTROL-BREAK key combination is hit
    { Not in Version 3.2 with Microsoft v5.1 but left in for others}
*****************************************************************************/
c_break ()
{
    fprintf (stderr, "Control-Break hit.  Program aborting ...\n");

    /*
     * if you created a file
     */
    if (TempName)
    {
	/*
	 * Make sure file is closed
	 */
	EndTempFile ();
	/*
	 * Delete temporary file
	 */
	unlink (TempName);
    }
    /*
     * ABORT sequence
     */
    return (0);
}

/*****************************************************************************/
void file_stats (filename)
    char *filename;
{
    extern struct find_t fblock;

    /*
     * read data into structure
     */
    _dos_findfirst (filename, 0, &fblock);

/* ------------------- Show date in standard format -------------------- */

    sprintf (file_date_buff, "%02d-%02d-%02d",

	     ((fblock.wr_date >> 5) & 0x0f),    /* month     */
	     (fblock.wr_date & 0x1f),           /* day       */
	     ((fblock.wr_date >> 9) + 80));	/* year      */

/* -----------------   Show time in 12 Hour Format   ------------------- */
    if ((fblock.wr_time >> 0x0b) <= 12)
    {
	sprintf (file_time_buff, "%02d:%02d",
		 (fblock.wr_time >> 0x0b),
		 ((fblock.wr_time >> 5) & 0x3f));	/* minutes   */
    }
    else
    {
	sprintf (file_time_buff, "%02d:%02d",
		 ((fblock.wr_time >> 0x0b) - 12),
		 ((fblock.wr_time >> 5) & 0x3f));	/* minutes   */
    }
/* ------------------     Decipher whether its AM or PM  ---------------- */
    if ((fblock.wr_time >> 0x0b) < 12)
	/*
	 * for AM
	 */
	strcat (file_time_buff, " am");
    else
	 /*
	  * for PM
	  */
	strcat (file_time_buff, " pm");
}

/*****************************************************************************/
int printer_code_init ()
/*
 * routine to initialize and select printer codes - called from initialize
 */
{
        printer = &(print_code[nprinter]);
    return (1);
}

/*****************************************************************************/
void normal_print ()
{
        fprintf (TempFile, "%s", printer->prt_normal);
}

/*****************************************************************************/
void emphasized_print_on ()
{
        fprintf (TempFile, "%s", printer->prt_emphasize_on);
}

/*****************************************************************************/
void emphasized_print_off ()
{
        fprintf (TempFile, "%s", printer->prt_emphasize_off);
}


/*****************************************************************************/
void print_double_width ()
{
        fprintf (TempFile, "%s", printer->prt_double_width_on);
}

/*****************************************************************************/
void compressed_print ()
{
        fprintf (TempFile, "%s", printer->prt_compress);
}

/*****************************************************************************/
void new_print_page ()
{
        fprintf (TempFile, "%s", printer->prt_new_page);
}

/*****************************************************************************/
void underline_on ()
{
        fprintf (TempFile, "%s", printer->prt_underline_on);
}


/*****************************************************************************/
void underline_off ()
{
        fprintf (TempFile, "%s", printer->prt_underline_off);
}

/*****************************************************************************/
void pspace (ns)
    int ns;
{
    register int i;

    for (i = 0; i < ns; i++)
	fprintf (TempFile, " ");
}

/*****************************************************************************/
char *strupper (str)
    char *str;
{
    register int c;
    register char *s = str;

    while (c = *s)
	*s++ = (char) toupper (c);
    return (str);
}

/*****************************************************************************/
char *strsave (s)
    char *s;
{
    char *p;

    if ((p = (char *) malloc (strlen (s) + 1)) != NULL)
	strcpy (p, s);
    return (p);
}

/*****************************************************************************/
void *strfree (s)
    char *s;
{
    free (s);
}

/*****************************************************************************/
/*
 * utility to strip leading and trailing blanks from a null terminated string
 */
char *strcompress (str)
    char *str;
{
    register char *s;

    s = str + strlen (str);
    *s--;
    while (*s == ' ')
	s--;
    *++s = '\0';
    while (*str == ' ')
	str++;
    return (str);
}

/*************************************************************************/
int CrossRefFunction (s)
    char *s;
{
    char *p,
        c;
    char refname[100];

    /*
     * assign pointer to function name string
     */
    p = refname;
    for (; *s; s++)
    {
	if (InXComment)
	    s = EndComment (s, &InXComment);
	if (InXString)
	    s = EndString (s, &InXString);
	switch (*s)
	{
	case '\'':
	    while ((c = *(++s)) && c != '\'');
	    if (*(s + 1) == '\'')
		/*
		 * special case
		 */
		s++;
	    break;

	case '"':
	    InXString = 1;
	    while ((c = *(++s)) && c != '\"');
	    if (c == '\"')
		InXString = 0;
	    break;
	case '/':
	    if (*(s + 1) == '*')
	    {
		InXComment = 1;
		s++;
		/***
		   * while( (c = *s++) )
		   * {
		   *    if( c == '*' && *s == '/' )
		   *    {
		   *       InXComment = 0;
		   *       break;
		   *    }
		   *  }
		   *  s--;
		   ***/
	    }
	    break;
	case '(':
	    *p = '\0';
	    write_xref (refname);
	    p = refname;
	    break;
	case '|':
	case '^':
	case '!':
	case '.':
	case '+':
	case '-':
	case '%':
	case '*':
	case '&':
	case '?':
	case '[':
	case ']':
	case '<':
	case '>':
	case ';':
	case ':':
	case '=':
	case ',':
	case ')':
	    /*
	    * start name over
	    */
	    p = refname;
	    break;

	default:
	    /*
	     * is it alpha,digit,'_*\t\n'?
	     */
	    if (isiechr (*s))
	    {
		/*
		* is it alpha,digit,'_' or ' '?
		*/
		if (isidchr (*s))
		    /*
		     * if it is store the byte for TOC
		     */
		    *p++ = *s;
		else
		    /*
		     * start over if '*\t\n' & nolp
		     */
		    p = refname;
		break;
	    }
	}			/* end of switch */
    }				/* end of for loop */
    return (1);
}

/*****************************************************************************/
/**
 * MJC
 * QUICK SCREEN CLEAR ROUTINE
 */
void clear_screen(void)
{
   _asm
      {
      xor   al,al
      xor   cx,cx
      mov   dh,24
      mov   dl,79
      mov   bh,7
      mov   ah,6
      int   10h
      }
}
